/* This file is part of SLB
   Copyright (C) 2011 Sergey Poznyakoff

   SLB is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   SLB is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with SLB.  If not, see <http://www.gnu.org/licenses/>. */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sysexits.h>
#include <syslog.h>
#include <errno.h>
#include <signal.h>

#include <regex.h>
#include <stdbool.h>
#include <stdio.h>
#include "c-ctype.h"
#include "obstack.h"
#include "progname.h"
#include "xalloc.h"
#include <inttypes.h>
#include "inttostr.h"

#include "gettext.h"

#define _(s) gettext(s)
#define N_(s) s

#include "grecs.h"

#include <net-snmp/net-snmp-config.h>
#include <net-snmp/net-snmp-includes.h>

extern int slb_sigtab[];
extern size_t slb_sigcount;
extern unsigned long slb_loop_serial;
extern time_t slb_loop_ts;

void set_signals(void (*handler) (int signo), int *sigv, int sigc);

void config_help(void);
void config_init(void);
void config_finish(void);

typedef int (*config_finish_hook_t)(void *);
void add_config_finish_hook(config_finish_hook_t fun, void *data);

extern int standalone_mode;
extern int foreground;
extern char *pidfile;
extern time_t wakeup_interval;


#define	SLB_DEBCAT_MAIN    0
#define	SLB_DEBCAT_EVAL    1
#define	SLB_DEBCAT_EGRAM   2
#define	SLB_DEBCAT_ELEX    3
#define	SLB_DEBCAT_SNMP    4
#define	SLB_DEBCAT_OUTPUT  5
#define SLB_DEBCAT_CFGRAM  6
#define SLB_DEBCAT_CFLEX   7
#define _SLB_DEBCAT_MAX    8

extern int debug_level[];

#define debug(cat, lev, s)				\
	do {						\
		if (debug_level[cat] >= (lev))		\
		       debug_printf s;			\
	} while(0)


enum node_type {
	node_unknown,
	node_variable,
	node_const,
	node_uminus,
	node_add,
	node_sub,
	node_mul,
	node_div,
	node_or,
	node_and,
	node_not,
	node_eq,
	node_ne,
	node_lt,
	node_le,
	node_gt,
	node_ge,
	node_cond,
	node_func,
	node_subexpr,
	node_pow
};

/* Maximum number of arguments to a built-in function */
#define SLB_MAX_ARGS 256

struct slb_node;
struct slb_expression;
struct slb_node_state;

typedef int (*slb_func_t) (struct slb_node_state **pstate,
			   struct slb_node *node,
			   int nargs, double *args,
			   double *res);

#define FUN_DAEMON_ONLY 0x01 /* Function runs only in daemon mode */

struct function {
	const char *name;   /* Function name */
	slb_func_t func;    /* Implementation */
	int minargs;        /* Min. number of arguments */
	int maxargs;        /* Max. number of arguments */
	int flags;          /* Flags */
	int min_eval;       /* Min. number of evaluations needed to
			       compute the result */
};

struct slb_node {
	enum node_type n_type;    /* Node type */
	struct slb_node *n_next;  /* Next node, if part of a list */
	int n_min_eval; /* Min. number of evaluations needed to
			   obtain the result */
	union {
		struct slb_node *n_arg[3];   /* arguments for arith. nodes */ 
		double n_const;              /* for node_const */ 
		struct slb_varref *n_var;    /* for node_variable */
		struct {
			slb_func_t func;         /* Function handling func. */ 
			struct slb_node *args;   /* List of actual arguments */
			size_t nargs;            /* Number of args */
		} n_func;                    /* for node_func */ 
		struct {
			char *name;
			struct slb_expression *expr;
		} n_expr;                    /* for node_subexpr */
	} v;
};

struct slb_node_list {
	struct slb_node *head;
	struct slb_node *tail;
	size_t count;
};


struct symtab;

struct syment {
	char *name;
};

typedef int (*symtab_enumerator_t)(void *sym, void *data);

const char *symtab_strerror(int rc);

void *symtab_lookup_or_install(struct symtab *st, void *key, int *install);
void symtab_clear(struct symtab *st);
struct symtab *symtab_create(size_t elsize, 
			     unsigned (*hash_fun)(void *, unsigned long),
			     int (*cmp_fun)(const void *, const void *),
			     int (*copy_fun)(void *, void *),
			     void *(*alloc_fun)(size_t),
			     void (*free_fun)(void *));
struct symtab *symtab_create_default(size_t elsize);

void symtab_destroy(struct symtab **pst);
int symtab_remove(struct symtab *st, void *elt);
int symtab_replace(struct symtab *st, void *ent, void **old_ent);
int symtab_enumerate(struct symtab *st, symtab_enumerator_t fun, void *data);

size_t symtab_count_entries(struct symtab *st);

struct slb_varref {
	struct syment vr_sym;
	struct slb_varinstance *vr_inst;
};

#define SLV_VIF_CONST   0x01
#define SLV_VIF_USED    0x02
#define SLV_VIF_SET     0x04

struct slb_varinstance {
	struct syment vi_sym;
	const char *vi_mib;
	oid vi_oid[MAX_OID_LEN];
	size_t vi_oidlen;
	double vi_value;
	int vi_flags;
	char *vi_extra;
};

struct slb_expression {
	const char *ex_name;      /* Expression name (if named) */
	const char *ex_text;      /* Actual expression text */
	int ex_fixed;             /* Is expression fixed up */
	grecs_locus_t ex_locus;   /* Location in the config. file */
	struct symtab *ex_vartab; /* Table of variables used in this expr. */
	struct slb_node *ex_tree; /* Compiled parse tree */
#define ex_min_eval ex_tree->n_min_eval
	int ex_eval_count;        /* Number of evaluations so far.
				     Always <= ex_min_eval. */
};

struct slb_macro {
	struct syment mac_sym;
	const char *mac_exp;
};


struct slb_varinstance *varinst_lookup(struct symtab *vit,
				       const char *name, size_t len,
				       int *install);
struct slb_varinstance *varinst_lookupz(struct symtab *vit,
					const char *name, int *install);
struct slb_varref *varref_lookup(struct symtab *vit,
				 const char *name, size_t len,
				 int *install);
struct slb_varref *varref_lookupz(struct symtab *vit,
				  const char *name, int *install);
struct slb_macro *macro_lookup(struct symtab *mt,
			       const char *name, size_t len,
			       int *install);
struct slb_macro *macro_lookupz(struct symtab *mt,
				const char *name, int *install);

void varinst_clear(struct symtab *vit);
int varinst_allset(struct symtab *vit);

struct symtab *create_oidtab(struct symtab *vit);
struct slb_varinstance *oidtab_lookup(struct symtab *oidtab,
				      oid *name, size_t name_length);


int register_expression(struct slb_expression *expr);
struct slb_expression *expression_lookupz(const char *name);
struct slb_expression *expression_lookup(const char *name, size_t len);
struct slb_expression *copy_expression(struct slb_expression *src,
				       struct symtab *vt);
int fixup_expression(struct slb_expression *expr, struct symtab *vit);
void bind_expression(struct slb_expression *expr, struct symtab *vit);

int compile_expression(const char *text, grecs_locus_t *locus,
		       struct slb_expression **pexpr);
int eval_expression(struct slb_expression *expr,
		    struct slb_node_state **pstate,
		    double *pnum);
struct slb_node *copy_node(struct slb_node *node, struct symtab *vartab);

void slb_node_state_init(struct slb_node_state *ns);

#define SLB_ASSERT_NE 0x01

struct slb_assertion
{
	oid name[MAX_OID_LEN];
	size_t length;
	int flags;
	char *value;
	char *text;
	grecs_locus_t locus;
	unsigned char *buf;
	size_t bufsize;
};

struct symtab *assertion_symtab_create(void);
struct slb_assertion *assertion_lookup(struct symtab *at, oid *name,
				       size_t namelen, int *install);


#define SLB_SRV_DISABLED 0x01
#define SLB_SRV_COMPUTED 0x02

struct slb_server {
	char *id;                  /* Server ID */
	grecs_locus_t locus;       /* Location in the config file */
	char *peer;                /* Server IP or host name */
	int port;                  /* Port it is listening on (not used yet) */
	time_t timeout;            /* SNMP timeout */
	unsigned retries;          /* Number of SNMP retries */
	char *community;           /* SNMP v1/v2c community */
	struct symtab *assertions; /* Assertions for this server */
	struct symtab *varinst;    /* Variable instances */
	struct symtab *oidtab;     /* A reverse index to the entries from
				      varinst */
	struct symtab *macros;     /* Defined macros */
	struct slb_node_state *state; /* Preserved node states for expression
					 evaluator */
	struct slb_expression *expr; /* Expression returning the server
					load "weight" */
	struct snmp_session *sess; /* SNMP session */        
	struct snmp_pdu *test_pdu; /* A PDU used only in test mode */
	double weight; /* Last computed server weight (relative load) */
	int flags;                 /* Server flags */
	struct slb_server *next;   /* Next server in the list */
};

extern struct slb_server *srv_head, *srv_tail;
extern size_t srv_count;
extern struct slb_server **srvtab;
extern size_t srvtab_count;

void snmploop(void);
void sort(void);
char *srvid(struct slb_server *srv);
struct slb_server *server_lookup(const char *id);
void server_process_pdu(int operation, struct slb_server *srv,
			struct snmp_pdu *pdu);


extern int log_to_stderr;
extern int syslog_facility;
extern const char *syslog_tag;
extern int syslog_include_prio;

void logger_setup(void);
void logmsg(int prio, const char *fmt, ...) SLB_PRINTFLIKE(2,3);
void logmsg_at_locus(int prio, grecs_locus_t *locus, const char *fmt, ...)
	SLB_PRINTFLIKE(3,4);
void debug_printf(const char *fmt, ...) SLB_PRINTFLIKE(1,2);

int string_to_syslog_facility(const char *str, int *pfacility);


extern size_t output_tail_count;
extern size_t output_head_count;
extern size_t output_suppress_count;
extern char *output_file;
extern char *output_format;
extern char *output_file_option;
extern char *begin_output_message;
extern char *end_output_message;

void slb_set_output_defaults(void);
void slb_open_output(void);
void slb_close_output(void);
void slb_check_output(void);

void slb_fmtout(void);

void collect_output_refs(void);
int bind_output_refname(struct slb_varinstance *vinst);
void undo_output_refs(void);
void report_unbound_output_refs(struct slb_server *srv);

void close_fds(fd_set *fdset);

int slb_test(const char *name);


